import matplotlib.pyplot as plt


def fig_ax(figsize=(15, 5), dpi=150):
    """Generate a (matplotlib) figure and ax objects with given size."""
    return plt.subplots(figsize=figsize, dpi=dpi)


def get_sum_of_cost(algo, n_bkps) -> float:
    """Calculate the sum of costs for the change points `bkps`.

    Utility function used to do penalized variable selection and obtain
    a final list of anomaly time points.
    """
    bkps = algo.predict(n_bkps=n_bkps)
    return algo.cost.sum_of_costs(bkps)



"""
This module, `kcp_v` (Kernel Change Point Visualisation), includes a function for visualization and
explainability. It extracts intervals around each detected change point, illustrating the
contribution of each sensor to the change point criterion. Each interval is bounded on the left by
either 0 or the previous change point location and on the right by either the next change point or
the time series end-point. The visualization highlights the change point's location with a vertical
dotted red line and displays the algorithm criterion as a blue curve.

Remarks:
1. For clarity in visualization, individual time series related to different change points are
visualized separately, resulting in one plot per detected change point.
2. The visualization function accommodates the algorithm's processing of a maximum of
"max_n_time_points" in each loop. If the total exceeds "max_n_time_points," the interval
calculation differs for the last change point found in each loop. Hence, "max_n_time_points" must
be consistent between the change point detection and visualization functions.
3. Only the linear kernel provides criterion lines for each individual time series that sum up to
the value of the global criterion line. Although designed for the linear kernel, the visualization
function can be used for other kernels (if the same kernel is used in the kcp_ds function).
4. #data: The input data array must be the exact same array used in the kcp_ds function.
"""

from typing import Sequence

import numpy as np
import matplotlib.pyplot as plt


def kcp_v(
    data: np.array,
    true_times: Sequence[int],
    detected_change_points: Sequence[int],
    interval_end_points: Sequence[int],
    n_legend: int = 10,
    save_plots: bool = False,
) -> None:
    """Visualize the post-hoc importance of each individual time series
    in the detection of each shared change point by the kernel change
    point detection algorithm.

    Args:
        data: array of dimension (number of time points) x (number of
            time-series) containing the data points.
        true_times: the dataset-dependent true time points (after any
            postprocessing steps required to interpolate data to a fixed
            time grid). This is required to make more meaningful plot
            outputs, or to obtain the actual true times of predicted
            change points. true_times should have the length of the
            first dimension of data.
        detected_change_points: the detected change points. It should be
            the first output of the kcp_ds function.
        interval_end_points: indices of the ends of intervals of length
            max_n_time_points. It should be the second output of kcpDS.
        n_legend: the number of "important" individual time series you
            want to be colorful in the final plot.
    """
    # Calculate the number of individual time series in parallel:
    n_time_series = np.shape(data)[1]

    # Stock the times that will be used to plot the criterion value. These times are halfway
    # between the actual data times, since when we say we detect a change point at time time, we
    # actually mean between t and the next data time, so for simplicity, we make this half-way
    # between the two:
    plot_times = [
        (true_times[i] + true_times[i + 1]) / 2 for i in range(len(true_times) - 1)
    ]

    # For each change point in detected_change_points:
    for i, _ in enumerate(detected_change_points):
        print("Change point number", i + 1, "\n")

        # Left edge Python index:
        L = -1 * (i == 0) + (detected_change_points[i - 1] - 1) * (i > 0)

        # Right edge Python index:
        # (Various cases to deal with here: (i) It's the last change point detected,
        # (ii) It's the last change point detected in a time block of length
        # "max_n_time_points", in which case the right edge is the end of the time
        # block, not the next change point)
        # (iii) Neither of the above:

        # If it is not the last change point detected:
        if i < len(detected_change_points) - 1:
            # If it is not the last change point in a block of "max_n_time_points"
            # time points: (to do this we need to look at the list "interval_end_points"
            # and check whether the (i+1)th and (i+2)th detected change points
            # are both smaller than the same number of elements in the list, i.e., whether
            # the were both found in the same loop of in kcpDS, or not)
            if np.sum(
                detected_change_points[i] < np.array(interval_end_points)
            ) == np.sum(detected_change_points[i + 1] < np.array(interval_end_points)):
                R = detected_change_points[i + 1] - 1
            else:
                # Extract the end point of the current interval from "interval_end_points", since
                # it comes BEFORE the next actually detected change point.
                n_greater_than = np.sum(
                    detected_change_points[i] > np.array(interval_end_points)
                )
                R = interval_end_points[int(n_greater_than)] - 1

        # If it is the last change point detected:
        else:
            R = np.shape(data)[0] - 1

        # Central change point:
        C = detected_change_points[i] - 1

        # Here we pull out the sub-dataset from times L to R, in order to examine it closer
        # and hopefully get some understanding on why, given this data, the change point was
        # detected at C.

        # arlot_crit is a list that will store the value of the kernel change point criterion for
        # each possible time at which we could have placed a change point between L and R. Due do
        # the fact that restricted to this interval, the solution for finding one (and one only)
        # change point will give the same change point as found by kcpDS globally (technical detail
        # due to how the criterion is calculated), we expect the MINIMUM of arlot_crit to be at C.
        # If it is not, then most likely Python indexing starting at 0 has probably created another
        # minor error somewhere. Thanks Python!
        arlot_crit = [0] * (R - L - 1)

        # Put our current data into an array:
        LR_data = np.transpose(data[(L + 1): (R + 1), :])

        # First calculate the diagonal term for the criterion for the LINEAR KERNEL:
        arlot_crit_diag = (1 / (R - L)) * np.sum(
            [np.sum(np.multiply(LR_data[:, j], LR_data[:, j])) for j in range(0, R - L)]
        )

        # Also, for later, calculate the criterion as if there were NO change points between L and R:
        arlot_crit_zero = arlot_crit_diag - (1 / (R - L)) * (1 / (R - L)) * np.sum(
            np.matmul(np.transpose(LR_data), LR_data)
        )

        # Now we cycle through the possible places to put one change point between L and R. This for
        # loop can be slow. There is no reason why it can't be parallelized if you are good enough at
        # Python to do so.
        for z in range(0, R - L - 1):
            C_z = L + 1 + z
            L_data_z = np.transpose(data[(L + 1): (C_z + 1), :])
            R_data_z = np.transpose(data[(C_z + 1): (R + 1), :])
            arlot_crit[z] = arlot_crit_diag - (1 / (R - L)) * (
                (1 / (C_z - L)) * (np.sum(np.matmul(np.transpose(L_data_z), L_data_z)))
                + (1 / (R - C_z))
                * (np.sum(np.matmul(np.transpose(R_data_z), R_data_z)))
            )

        # Now we must do exactly the same for each individual time series. Basically it is the
        # same code as just above, but for each time series, one at a time.

        # arlot_crit_sense_array has the same job as arlot_crit above, except it requires
        # one row per time series:
        arlot_crit_sense_array = np.zeros((n_time_series, R - L - 1))

        # arlot_crit_zero_zz has the same function as arlot_crit_zero above.
        arlot_crit_zero_zz = [0] * n_time_series

        # Here we cycle through the individual time series, one by one. There is no
        # barrier to parallelizing this.
        for m in range(0, n_time_series):
            arlot_crit_sense = [0] * (R - L - 1)
            LR_data_zz = np.transpose(data[(L + 1): (R + 1), m]).reshape(1, -1)

            # First calculate the diagonal:
            arlot_crit_diag_zz = (1 / (R - L)) * np.sum(
                [
                    np.sum(np.multiply(LR_data_zz[0, j], LR_data_zz[0, j]))
                    for j in range(0, R - L)
                ]
            )

            # Also calculate the criterion as if there were no change points, for later use.
            arlot_crit_zero_zz[m] = arlot_crit_diag_zz - (1 / (R - L)) * (
                1 / (R - L)
            ) * (np.sum(np.matmul(np.transpose(LR_data_zz), LR_data_zz)))

            # This inner for loop could also be parallelized.
            for z in range(0, R - L - 1):
                C_z = L + 1 + z
                L_data_zz = np.transpose(data[(L + 1): (C_z + 1), m]).reshape(1, -1)
                R_data_zz = np.transpose(data[(C_z + 1): (R + 1), m]).reshape(1, -1)
                arlot_crit_sense[z] = arlot_crit_diag_zz - (1 / (R - L)) * (
                    (1 / (C_z - L))
                    * (np.sum(np.matmul(np.transpose(L_data_zz), L_data_zz)))
                    + (1 / (R - C_z))
                    * (np.sum(np.matmul(np.transpose(R_data_zz), R_data_zz)))
                )

            arlot_crit_sense_array[m, :] = arlot_crit_sense

        plt.figure(figsize=(15, 10))

        plot_points = plot_times[(L + 1): R]

        plt.plot(plot_points, (arlot_crit - arlot_crit_zero), label="Combined")

        plt.axvline(plot_times[(detected_change_points[i] - 1)], ls="--", c="red", lw=1)

        plt.xlabel("Time")
        plt.ylabel("Kernel change-point criterion")

        # Here we want to look at the value of the criterion for each individual time series
        # at C, and rank them from smallest to largest. We will then plot the first few smallest
        # on the same plot as the global criterion.
        val_at_global_min = [
            arlot_crit_sense_array[x, C - L - 1] - arlot_crit_zero_zz[x]
            for x in range(n_time_series)
        ]

        sorted_indices = np.argsort(val_at_global_min)

        # The ones we keep
        keep_indices = sorted_indices[0:n_legend]

        for m in range(0, n_time_series):
            if m in keep_indices:
                plt.plot(
                    plot_times[(L + 1): R],
                    arlot_crit_sense_array[m, :] - arlot_crit_zero_zz[m],
                    label=m,
                )
            else:
                plt.plot(
                    plot_times[(L + 1): R],
                    arlot_crit_sense_array[m, :] - arlot_crit_zero_zz[m],
                    color="black",
                )
        plt.legend(title="Important time series")

        if save_plots:
            plt.savefig(f"plot{i}")

        plt.show()



"""
This module, `kcp_ds` (Kernel Change Point Detect Select), encapsulates functions from the Ruptures
package and includes a function for penalized variable selection.

Specific details and remarks:
1. The function outputs Python indices of the columns of "data" corresponding to the detected
change points. Real-world times associated with the data are not considered in this function.
2. The ruptures class KernelCPD has a minimum allowed segment size (min_size). It is currently set
to 2 by default and is not an input parameter for kcp_ds.
3. The default kernel is "linear," with options for "cosine" and "rbf" kernels. To cover all bases,
an input params is required, which defaults to None but can be modified for the rbf kernel (e.g.,
params = {"gamma": value_of_gamma}).
4. Due to memory constraints, the algorithm processes offline time-series data longer than max_n_time_points
in segments of length max_n_time_points, and then outputs the full set of change points.

Note: Please handle real-world times associated with the data independently as this module focuses
on change point detection and variable selection.
"""

from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple
import math

import numpy as np
import ruptures as rpt
from sklearn.linear_model import LinearRegression

from kcpdi.utils import get_sum_of_cost


def kcp_ds(
    data: np.array,
    kernel: Literal["linear", "cosine", "rbf"] = "linear",
    params: Optional[Dict[str, Any]] = None,
    max_n_time_points: int = 2000,
    min_n_time_points: int = 10,
    expected_frac_anomaly: float = 1 / 1000,
) -> Tuple[List[int], List[int]]:
    """Return a set of important change points change points after
    running kernel change point detection followed by penalized model
    selection using the slope heuristic method.

    Since the criterion is responding to noise/overfitting, penalized
    variable selection is performed in order to obtain a final decision
    as to how many of the detected change points are truly significant
    and not simply adjusted noise.

    Args:
        data: array of dimension (number of time points) x (number of
            time-series) containing the data points.
        kernel: Kernel of the kernet change_point detection. If it is
            rbf, params must include a "gamma" parameters, with a
            positive real value.
        params: parameters for the kernel instance
        max_n_time_points: maximum size (max_n_time_points) x
            (max_n_time_points) of matrices expected to be processed
            quickly by the computer system.
        min_n_time_points: minimum number of time points in the current
            dataset for which it makes sense to run the detection
            algorithm again. If there are fewer than min_n_time_points
            points in the dataset, no computations will be run and the
            outputs will be empty.
        expected_frac_anomaly: This parameter encodes prior knowledge
            of users as to how often anomalies might occur. Results can
            be quite dependent on the choice of this parameter, so
            choose carefully!
    """
    if max_n_time_points < min_n_time_points:
        raise ValueError("max_n_time_points should be greater than min_n_time_points.")

    if expected_frac_anomaly > 1 / 2:
        raise ValueError("expected_frac_anomaly should be at most 1/2.")

    n_total_time_points = np.shape(data)[0]

    current_data = np.copy(data)  # new variable for internal manipulation

    n_current_time_points = np.shape(current_data)[0]
    curr_n_remaining_t_points = n_current_time_points

    # Initialise an empty list to store the end points of the intervals of length max_n_time_points
    # (except for the last interval, which could be shorter). This variable will be required for
    # the explainability function that can be run after the current function:
    interval_end_points = []

    current_zero = 0  # remember the original Python indices

    detected_change_points = []

    while curr_n_remaining_t_points > max_n_time_points:
        # Get the block of data to plug into the Ruptures setting:
        data_use = current_data[:max_n_time_points, :]

        # Store the true index of the time point at the end of this interval, which will be needed
        # in the explainability function later:
        interval_end_points.append(current_zero + max_n_time_points)

        detected_change_points, next_ch_pts = _detect_changepoints_from_kernelcpd(
            kernel=kernel,
            params=params,
            data_use=data_use,
            expected_frac_anomaly=expected_frac_anomaly,
            current_zero=current_zero,
            detected_change_points=detected_change_points,
        )

        # Concatenate any current data rows left after the last change point found, to the
        # remaining data (if there is any left):
        current_data = data[(next_ch_pts[-1] - 1):, :]

        current_zero = next_ch_pts[-1] - 1

        n_current_time_points = np.shape(current_data)[0]
        curr_n_remaining_t_points = n_current_time_points

    # Deal with any residual block of data that is shorter than (or equal to)
    # max_n_time_points but still larger than min_n_time_points:
    if (curr_n_remaining_t_points <= max_n_time_points) & (
        curr_n_remaining_t_points >= min_n_time_points
    ):
        # Store the true index of the time point at the end of this interval, which will be needed
        # in the explainability function later. Unlike in the above while loop, here the time point
        # at the end of the interval is the last time point in the dataset:
        interval_end_points.append(n_total_time_points)

        detected_change_points, next_ch_pts = _detect_changepoints_from_kernelcpd(
            kernel=kernel,
            params=params,
            data_use=current_data,
            expected_frac_anomaly=expected_frac_anomaly,
            current_zero=current_zero,
            detected_change_points=detected_change_points,
        )

    # We need to remove any "detected" change-points which were actually ends of intervals.
    # First we add 1 to each detected_change_point:
    temp_detec = [
        detected_change_points[i] + 1 for i in range(len(detected_change_points))
    ]
    # Now look for matches with values in "interval_end_points"
    index_matches = [
        i
        for i in range(len(detected_change_points))
        if temp_detec[i] in interval_end_points
    ]
    detected_change_points = [
        i for j, i in enumerate(detected_change_points) if j not in index_matches
    ]

    return detected_change_points, interval_end_points


def _detect_changepoints_from_kernelcpd(
    kernel: Literal["linear", "cosine", "rbf"],
    params: Optional[Dict[str, Any]],
    data_use: np.array,
    expected_frac_anomaly: float,
    current_zero: int,
    detected_change_points: List[int],
) -> Tuple[List[int], List[int]]:
    """Apply kernel change point detection and update
    detected_change_points.

    Auxiliary function to kcp_ds. Update the detected_change_points
    variable in place.

    Args:
        kernel: Kernel of the kernet change_point detection. If it is
            rbf, params must include a "gamma" parameters, with a
            positive real value.
        params: parameters for the kernel instance
        data_use: chunk of data on which to apply the detection
        expected_frac_anomaly: This parameter encodes prior knowledge
            of users as to how often anomalies might occur. Results can
            be quite dependent on the choice of this parameter, so
            choose carefully!
        current_zero: index of the first point of the current chunk of
            the dataset
        detected_change_points: all breakpoints detected so far

    Returns:
        The list of all points detected so far.
        The list of points newly detected during the last iteration.
    """
    # Associate the data with the class KernelCPD from the Ruptures library:
    algo = rpt.KernelCPD(kernel=kernel, params=params)
    algo.fit(data_use)

    # Get a value for the maximum number of change-points to look for; essentially,
    # this is set as = 2 * fraction of expected anomalies * number of current time points.
    n_bkps_max = max(
        1, int(2 * np.floor(expected_frac_anomaly * np.shape(data_use)[0]))
    )

    array_of_n_bkps = np.arange(1, n_bkps_max + 1)
    algo_costs = [
        get_sum_of_cost(algo=algo, n_bkps=n_bkps) for n_bkps in array_of_n_bkps
    ]  # value of the criterion for when we have chosen to stick with 1 up to n_bkps_max points
    # Since each time we add an extra change point, the criterion will never decrease:
    # algo_costs is a non-increasing list.

    # For the linear kernel we want to concatenate onto the FRONT of this list the value of the
    # criterion when there are ZERO change-points. The problem is that the ruptures package
    # assumes that there is at least one change-point, while the model selection of Arlot et al
    # (2019) includes the case when there are ZERO change-points. We emphasize that the matrix
    # multiplication below is only for the linear kernel.
    linear_prod = np.dot(data_use, data_use.T)
    diag_sum = np.sum([linear_prod[i, i] for i in range(np.shape(data_use)[0])])
    crit_linear_zero_cp = diag_sum - np.mean(linear_prod)
    # Concatenate this value onto the FRONT of the algo_costs list:
    algo_costs.insert(0, crit_linear_zero_cp)

    n_bkps_pen = _model_select_n_bp(
        data=data_use,
        algo_costs=algo_costs,
        array_of_n_bkps=array_of_n_bkps,
        n_bkps_max=n_bkps_max,
    )  # penalized selection

    # Extract the solution:
    if n_bkps_pen > 0:
        next_ch_pts = algo.predict(n_bkps=n_bkps_pen)
        # Remove the end point, which itself is not an endpoint:
        del next_ch_pts[-1]
    else:
        # If there were NO change-points detected, by convention we shall
        # instead output the end-point index of the interval AS IF it were
        # a change-point. Note that this (index + 1) was also already stored in
        # the variable "interval_end_points" so we will be able to weed out
        # this change-point at the end for the final list and for the
        # eventual visualization steps.
        next_ch_pts = algo.predict(n_bkps=1)
        next_ch_pts = [next_ch_pts[-1] - 1]

    # Add the current zero index (in order to get the true row/time point in the data):
    next_ch_pts = [x + current_zero for x in next_ch_pts]

    # Concatenate these with any found in previous loops:
    detected_change_points = detected_change_points + list(next_ch_pts)

    return detected_change_points, next_ch_pts


def _model_select_n_bp(
    data: np.array,
    algo_costs: Sequence[float],
    array_of_n_bkps: Sequence[int],
    n_bkps_max: int,
) -> int:
    """Calculate the number of change points after model selection using
    the slope heuristic method.

    This function takes the output set of candidate change points from
    the Ruptures library method and refines them to a subset that is
    likely to represent signal rather than (over)fitted noise. It
    implements the penalized variable selection method described
    in Arlot et al. (2019).

    Args:
        data: signal data points in which to calculate the optimal
            number of change points
        algo_costs: the (non-penalized) costs associated with each
            number of change points
        array_of_n_bkps: array listing the number of change points to be
            considered
        n_bkps_max: maximum number of change points to be detected

    Returns:
        The number of change points after penalized variable selection
        using the slope heuristic method in Arlot et al. (2019).
    """
    lower_D = math.ceil(0.6 * n_bkps_max)

    X1 = [
        (1 / np.shape(data)[0]) * math.log(math.comb(np.shape(data)[0] - 1, D - 1))
        for D in np.arange(lower_D, n_bkps_max + 1)
    ]
    X2 = [(1 / np.shape(data)[0]) * D for D in np.arange(lower_D, n_bkps_max + 1)]

    X = np.transpose(np.array([X1, X2]))
    y = np.array(
        [
            (1 / np.shape(data)[0]) * algo_costs[D - 1]
            for D in np.arange(lower_D, n_bkps_max + 1)
        ]
    )
    #
    # Linear regression:
    algo_reg = LinearRegression().fit(X, y)
    s1_hat = algo_reg.coef_[0]
    s2_hat = algo_reg.coef_[1]

    ###########################################################

    # Final parameters:
    c1 = -2 * s1_hat
    c2 = -2 * s2_hat

    # Use these two parameters to do the model selection in Arlot et al. (2019):

    algo_pen = [
        algo_costs[i - 1]
        + c1 * math.log(math.comb(np.shape(data)[0] - 1, i - 1))
        + c2 * i
        for i in array_of_n_bkps
    ]

    min_val = min(algo_pen)
    all_min_indices = [i for i, x in enumerate(algo_pen) if x == min_val]

    # There could be more than one items in this object. We choose the smallest for parcimony
    # reasons in this (unlikely) event:
    min_index = all_min_indices[0]

    n_bkps_pen = min_index

    return n_bkps_pen



"""
This module, `kcp_ss_learner` (Kernel Change Point Sample Scorer Learner), includes a wrapper to
perform anomaly scoring as understood in the TADkit formalism, based on this library's change-point
detection methodology.

The kernel change-point detection algorithm outputs a final list of time indices at which it thinks
that "something happened".

Other methods output a score at __all__ time indices, where the higher the score is, the more we
believe that "something happened there".

In order to allow the kernel change-point method to be integrated into a Python package based
around score samples, we have implemented a function __kcp_ss__ which takes the __kcp_ds__ output
list of change-points and turns them into scores at __all time indices__.

Remember that the kernel change-point detection algorithm truly believes that it has found the true
and only set of change-points. However, due to random noise, it could have been that a true
change-point was immediately before or immediately after a detected change-point, i.e., 1 or 2 or
maybe even 3 time indices before or after.

We give a score of 1 at each detected change-point. Other points are assigned a score that
decreases as they move away from the detected change-point, in a way parameterized by a decay
parameter (gamma).
"""
from typing import Any, Dict, List, Literal, Optional, Sequence

from sklearn.base import BaseEstimator, OutlierMixin

from kcpdi import kcp_ds

import numpy as np


class KcpLearner(BaseEstimator, OutlierMixin):
    required_properties = []

    params_description = {
        "decay_param": {
            "description": "How fast the score decays around the change-points.",
            "family": "postprocessing",
            "value_type": "log_range",
            "log_start": -2,
            "log_step": 0.1,
            "log_stop": 2,
            "default": 1,
        }
    }

    def __init__(
            self,
            kernel: Literal["linear", "cosine", "rbf"] = "linear",
            params: Optional[Dict[str, Any]] = None,
            max_n_time_points: int = 2000,
            min_n_time_points: int = 10,
            expected_frac_anomaly: float = 1 / 1000,
            decay_param: float = 1.,
    ):
        self.kernel = kernel
        self.params = params
        self.max_n_time_points = max_n_time_points
        self.min_n_time_points = min_n_time_points
        self.expected_frac_anomaly = expected_frac_anomaly
        self.decay_param = decay_param

    def fit(self, X, y=None):
        return self

    @staticmethod
    def _kcp_ss(
            detected_change_points: Sequence[int],
            n_time_points: int,
            decay_param: float = 1.,
    ) -> List[float]:
        """Take the output list of detected change-points from the
        function kcp_ds() and transform them into a score at each time
        point from 0 (low probability of an anomaly at the time point)
        to 1 (high probability of an anomaly at the time point).

        Args:
            detected_change_points: The detected change-points. It
                should be the first output of the kcp_ds function.
            n_time_points: The number of time points in the input data.
                Usually this will be something like: np.shape(data)[0].
            decay_param: Parameter that defines how fast the score
                decays around the change-points. The larger this value,
                the faster it decays.
        """
        if decay_param < 0:
            raise ValueError(
                "decay_param has to be positive."
            )

        # Create list of 0s of length n_time_points
        kcp_scores = [0] * n_time_points

        if not len(detected_change_points):
            return kcp_scores

        for i in range(n_time_points):

            # Cases:
            # 1) i is in the list of change-point indices
            # 2) i is before the first change-point index
            # 3) i is between two change-point indices
            # 4) i is after the last change-point index

            # Create a local version of the input variable detected_change_points
            # and append i to it:
            local_detected_change_points = detected_change_points.copy()
            local_detected_change_points.append(i)

            # A couple of useful values:
            sum_first = sum(i < j for j in detected_change_points)
            sum_last = sum(i > j for j in detected_change_points)

            # Case (1)
            if len(set(local_detected_change_points)) == len(detected_change_points):
                kcp_scores[i] = 1

            # Case (2)
            elif sum_first == len(detected_change_points):
                # We only have the distance to the first change-point index to deal with:
                curr_dist = abs(i - detected_change_points[0])
                kcp_scores[i] = (1 / 2) ** (curr_dist * decay_param)

            # Case (4)
            elif sum_last == len(detected_change_points):
                # We only have the distance to the last change-point index to deal with:
                curr_dist = abs(i - detected_change_points[-1])
                kcp_scores[i] = (1 / 2) ** (curr_dist * decay_param)

            # Case (3)
            else:
                n_left = sum(i > j for j in detected_change_points)
                curr_dist_left = abs(i - detected_change_points[n_left - 1])
                curr_dist_right = abs(i - detected_change_points[n_left])
                kcp_scores[i] = (1 / 2) * (
                    (1 / 2) ** (curr_dist_left * decay_param) +
                    (1 / 2) ** (curr_dist_right * decay_param)
                )

        return kcp_scores

    def score_samples(self, X):
        if hasattr(X, "values"):
            # small hack to handle DataFrame s
            X = X.values
        detected_change_points, _ = kcp_ds(
            data=X,
            kernel=self.kernel,
            params=self.params,
            max_n_time_points=self.max_n_time_points,
            min_n_time_points=self.min_n_time_points,
            expected_frac_anomaly=self.expected_frac_anomaly,
        )

        return np.array([- score for score in self._kcp_ss(
            detected_change_points=detected_change_points,
            n_time_points=len(X),
            decay_param=self.decay_param
        )])
